@@ -17,7 +17,8 @@ from streamflow.workflow.step import BaseStep
 from jupyter_workflow.streamflow import executor
 from jupyter_workflow.streamflow.combinator import JupyterCartesianProductInputCombinator
 from jupyter_workflow.streamflow.command import JupyterCommand, JupyterCommandToken
-from jupyter_workflow.streamflow.token_processor import FileTokenProcessor, NameTokenProcessor, ControlTokenProcessor
+from jupyter_workflow.streamflow.token_processor import FileTokenProcessor, NameTokenProcessor, ControlTokenProcessor, \
+    OutputLogTokenProcessor
 
 
 def _build_dependencies(workflow: Workflow, in_step: Step) -> None:
@@ -55,31 +56,31 @@ def _extract_dependencies(cell_name: Text,
     return list(visitor.deps)
 
 
-def _get_combinator_from_schema(step: Step,
+def _get_combinator_from_scatter(step: Step,
                                 scatter_ports: MutableMapping[Text, InputPort],
-                                scatter_schema: Optional[MutableMapping[Text, Any]] = None) -> InputCombinator:
-    scatter_method = scatter_schema.get('method', 'cartesian')
+                                scatter: Optional[MutableMapping[Text, Any]] = None) -> InputCombinator:
+    scatter_method = scatter.get('method', 'cartesian')
     combinator_name = utils.random_name()
     if scatter_method == 'cartesian':
         combinator = JupyterCartesianProductInputCombinator(name=combinator_name, step=step)
     else:
         combinator = DotProductInputCombinator(name=combinator_name, step=step)
-    for entry in scatter_schema['items'] or {}:
-        if isinstance(entry, str):
-            combinator.ports[entry] = scatter_ports[entry]
-        else:
-            inner_combinator = _get_combinator_from_schema(
-                step=step,
-                scatter_ports=scatter_ports,
-                scatter_schema=entry)
-            combinator.ports[inner_combinator.name] = inner_combinator
+    if scatter:
+        for entry in scatter.get('items') or []:
+            if isinstance(entry, str):
+                combinator.ports[entry] = scatter_ports[entry]
+            else:
+                inner_combinator = _get_combinator_from_scatter(
+                    step=step,
+                    scatter_ports=scatter_ports,
+                    scatter=entry)
+                combinator.ports[inner_combinator.name] = inner_combinator
     return combinator
 
 
 def _get_input_combinator(step: Step,
-                          scatter_inputs: Optional[Set[Text]] = None,
-                          scatter_schema: Optional[Any] = None) -> InputCombinator:
-    scatter_inputs = scatter_inputs or set()
+                          scatter: Optional[Any] = None) -> InputCombinator:
+    scatter_inputs = _get_scatter_inputs(scatter)
     # If there are no scatter ports in this step, create a single DotProduct combinator
     if not [n for n in scatter_inputs]:
         input_combinator = DotProductInputCombinator(name=utils.random_name(), step=step)
@@ -97,10 +98,10 @@ def _get_input_combinator(step: Step,
                 scatter_ports[port_name] = port
                 del other_ports[port_name]
         # Choose the right combinator for the scatter ports, based on the scatter method property
-        scatter_combinator = _get_combinator_from_schema(
+        scatter_combinator = _get_combinator_from_scatter(
             step=step,
             scatter_ports=scatter_ports,
-            scatter_schema=scatter_schema)
+            scatter=scatter)
         cartesian_combinator.ports[scatter_combinator.name] = scatter_combinator
         # Create a CartesianProduct combinator between the scatter ports and the DotProduct of the others
         if other_ports:
@@ -113,14 +114,25 @@ def _get_input_combinator(step: Step,
 
 def _get_scatter_inputs(scatter_schema: Optional[MutableMapping[Text, Any]]) -> Set:
     scatter_inputs = set()
-    for entry in scatter_schema['items'] or {}:
-        if isinstance(entry, str):
-            scatter_inputs.add(entry)
-        else:
-            scatter_inputs.update(_get_scatter_inputs(entry))
+    if scatter_schema:
+        for entry in scatter_schema.get('items') or []:
+            if isinstance(entry, str):
+                scatter_inputs.add(entry)
+            else:
+                scatter_inputs.update(_get_scatter_inputs(entry))
     return scatter_inputs
 
 
+def _substitute_in_combinator(combinator: InputCombinator, port: InputPort):
+    if port.name in combinator.ports:
+        combinator.ports[port.name] = port
+        return combinator
+    combinators = filter(lambda c: isinstance(c, InputCombinator), combinator.ports.values())
+    for comb in combinators:
+        combinator.ports[comb.name] = _substitute_in_combinator(comb, port)
+    return combinator
+
+
 class NamesStack(object):
 
     def __init__(self):
@@ -305,6 +317,52 @@ class JupyterNotebookTranslator(object):
     def __init__(self, context: StreamFlowContext):
         self.context: StreamFlowContext = context
 
+    async def optimize_scatter(self,
+                               workflow: Workflow) -> None:
+        # Collect gather ports
+        gather_ports = set()
+        for s in workflow.steps.values():
+            for p in s.output_ports.values():
+                if isinstance(p, GatherOutputPort):
+                    gather_ports.add(posixpath.join(s.name, p.name))
+        # Remove gather ports that are workflow output ports
+        for output_port in workflow.output_ports.values():
+            gather_ports.discard(posixpath.join(output_port.step.name, output_port.name))
+        # Remove gather ports that do not have only scatter ports in their successors
+        for step in workflow.steps.values():
+            for port in step.input_ports.values():
+                if port.dependee:
+                    dependee = posixpath.join(port.dependee.step.name, port.dependee.name)
+                    if dependee in gather_ports:
+                        if not isinstance(port, ScatterInputPort):
+                            gather_ports.discard(dependee)
+        # Collect scatter ports
+        scatter_ports = set()
+        for s in workflow.steps.values():
+            for p in s.input_ports.values():
+                if isinstance(p, ScatterInputPort):
+                    scatter_ports.add(p)
+        # If port depends on a survived gather port, perform optimization
+        optimized_out_ports = {}
+        for port in scatter_ports:
+            if port.dependee:
+                dependee = posixpath.join(port.dependee.step.name, port.dependee.name)
+                if dependee in gather_ports:
+                    in_port = DefaultInputPort(name=port.name, step=port.step)
+                    in_port.token_processor = port.token_processor
+                    port.step.input_ports[port.name] = in_port
+                    port.step.input_combinator = _substitute_in_combinator(
+                        combinator=port.step.input_combinator,
+                        port=in_port)
+                    if dependee in optimized_out_ports:
+                        in_port.dependee = optimized_out_ports[dependee]
+                    else:
+                        out_port = DefaultOutputPort(name=port.dependee.name, step=port.dependee.step)
+                        out_port.token_processor = port.dependee.token_processor
+                        port.dependee.step.output_ports[port.dependee.name] = out_port
+                        in_port.dependee = out_port
+                        optimized_out_ports[dependee] = out_port
+
     async def translate_cell(self,
                              cell: JupyterCell,
                              metadata: Optional[MutableMapping[Text, Any]],
@@ -331,7 +389,7 @@ class JupyterNotebookTranslator(object):
         # Process cell inputs
         cell_inputs = metadata['step'].get('in', [])
         input_tokens = {}
-        scatter_inputs = _get_scatter_inputs(metadata['step'].get('scatterSchema'))
+        scatter_inputs = _get_scatter_inputs(metadata['step'].get('scatter'))
         for element in cell_inputs:
             # If is a string, it refers to the name of a variable
             if isinstance(element, Text):
@@ -382,7 +440,10 @@ class JupyterNotebookTranslator(object):
             input_names = _extract_dependencies(cell.name, cell.compiler, cell.code)
             for name in input_names:
                 if name not in step.input_ports:
-                    port = DefaultInputPort(name=name, step=step)
+                    if name in scatter_inputs:
+                        port = ScatterInputPort(name=name, step=step)
+                    else:
+                        port = DefaultInputPort(name=name, step=step)
                     port.token_processor = NameTokenProcessor(
                         port=port,
                         compiler=cell.compiler,
@@ -440,14 +501,19 @@ class JupyterNotebookTranslator(object):
                     # Register step port
                     step.output_ports[name] = output_port
         # Add output log port
-        output_log_port = DefaultOutputPort(name=executor.CELL_OUTPUT, step=step)
-        output_log_port.token_processor = ControlTokenProcessor(port=output_log_port)
+        if scatter_inputs:
+            output_log_port = GatherOutputPort(
+                name=executor.CELL_OUTPUT,
+                step=step,
+                merge_strategy=lambda values: sorted(values, key=lambda t: int(posixpath.basename(t.tag))))
+        else:
+            output_log_port = DefaultOutputPort(name=executor.CELL_OUTPUT, step=step)
+        output_log_port.token_processor = OutputLogTokenProcessor(port=output_log_port)
         step.output_ports[executor.CELL_OUTPUT] = output_log_port
         # Set input combinator for the step
         step.input_combinator = _get_input_combinator(
             step=step,
-            scatter_inputs=scatter_inputs,
-            scatter_schema=metadata['step'].get('scatterSchema'))
+            scatter=metadata['step'].get('scatter'))
         # Create the command to be executed remotely
         step.command = JupyterCommand(
             step=step,
@@ -466,13 +532,11 @@ class JupyterNotebookTranslator(object):
         for cell in notebook.cells:
             step = await self.translate_cell(
                 cell=cell,
-                metadata={**cell.metadata, **notebook.metadata},
+                metadata={**notebook.metadata, **cell.metadata},
                 autoawait=notebook.autoawait)
             _build_dependencies(workflow, step)
             workflow.steps[step.name] = step
-        # Extract workflow outputs
-        last_step = workflow.steps[notebook.cells[-1].name]
-        for port_name, port in last_step.output_ports.items():
-            workflow.output_ports[last_step.name] = port
+        # Apply rewrite rules
+        await self.optimize_scatter(workflow)
         # Return the final workflow object
         return workflow
